# Copyright 2022 Lawrence Livermore National Security, LLC and other
# Thicket Project Developers. See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT

import seaborn as sns

from ..utils import verify_thicket_structures


def display_heatmap(thicket, columns=None, **kwargs):
    """Display a heatmap which contains a full list of nodes and user passed columns.
    Columns must be from the aggregated statistics table.

    Arguments:
        thicket (thicket): Thicket object
        columns (list): List of hardware/timing metrics from aggregated statistics table
            to display. Note, if using a columnar joined thicket a list of tuples must
            be passed in with the format (column index, column name).

    Returns:
        (matplotlib Axes): Object for managing heatmap plot.
    """
    if columns is None:
        raise ValueError(
            "To see a list of valid columns, run Thicket.get_perf_columns()."
        )

    verify_thicket_structures(
        thicket.statsframe.dataframe, index=["node"], columns=columns
    )

    # thicket object without columnar index
    if thicket.dataframe.columns.nlevels == 1:
        thicket.statsframe.dataframe.index = thicket.statsframe.dataframe.index.map(str)
        ax = sns.heatmap(thicket.statsframe.dataframe[columns], **kwargs)

        return ax
    # columnar joined thicket object
    else:
        thicket.statsframe.dataframe.index = thicket.statsframe.dataframe.index.map(str)

        initial_idx = columns[0][0]
        cols = [columns[0][1]]
        for i in range(1, len(columns)):
            if initial_idx != columns[i][0]:
                raise ValueError(
                    "Columns specified as tuples must have the same column index (first element)."
                )
            else:
                cols.append(columns[i][1])

        ax = sns.heatmap(
            thicket.statsframe.dataframe[initial_idx][cols], **kwargs
        ).set_title(initial_idx)

        return ax



# Copyright 2022 Lawrence Livermore National Security, LLC and other
# Thicket Project Developers. See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT

import pandas as pd
import numpy as np

from ..utils import verify_thicket_structures


def calc_boxplot_statistics(thicket, columns=[], quartiles=[0.25, 0.5, 0.75], **kwargs):
    """Calculate boxplots lowerfence, q1, q2, q3, iqr, upperfence, and outliers for each
    node in the performance data table.

    Designed to take in a thicket, and append one or more columns to the aggregated
    statistics table for the boxplot five number summary calculations for each node.

    Arguments:
        thicket (thicket): Thicket object
        columns (list): List of columns to calculate boxplot metrics on. Note, if using
            a columnar joined thicket a list of tuples must be passed in with the format
            (column index, column name).
        quartiles (list): List containing three values between 0 and 1 to cut the
            distribution into equal probabilities.
    """
    if len(quartiles) != 3:
        raise ValueError(
            "The argument quartiles takes a list containing three numbers between 0 and 1."
        )

    if len(columns) == 0:
        raise ValueError(
            "To see a list of valid columns, please run Thicket.get_perf_columns()."
        )

    verify_thicket_structures(
        thicket.dataframe, index=["node", "profile"], columns=columns
    )

    q_list = str(tuple(quartiles))

    # thicket object without columnar index
    if thicket.dataframe.columns.nlevels == 1:
        for col in columns:
            boxplot_dict = {
                col + "_q1" + q_list: [],
                col + "_q2" + q_list: [],
                col + "_q3" + q_list: [],
                col + "_iqr" + q_list: [],
                col + "_lowerfence" + q_list: [],
                col + "_upperfence" + q_list: [],
                col + "_outliers" + q_list: [],
            }

            for node in pd.unique(thicket.dataframe.reset_index()["node"].tolist()):
                values = thicket.dataframe.loc[node][col].tolist()

                q = np.quantile(values, quartiles)
                q1 = q[0]
                median = q[1]
                q3 = q[2]

                iqr = q3 - q1
                lower_fence = q1 - (1.5 * iqr)
                upper_fence = q3 + (1.5 * iqr)

                boxplot_dict[col + "_q1" + q_list].append(q1)
                boxplot_dict[col + "_q2" + q_list].append(median)
                boxplot_dict[col + "_q3" + q_list].append(q3)
                boxplot_dict[col + "_iqr" + q_list].append(iqr)
                boxplot_dict[col + "_lowerfence" + q_list].append(lower_fence)
                boxplot_dict[col + "_upperfence" + q_list].append(upper_fence)

                profile = []
                for i in range(0, len(values)):
                    if values[i] > upper_fence or values[i] < lower_fence:
                        profile.append(
                            thicket.dataframe.loc[node].reset_index()["profile"][i]
                        )
                    else:
                        continue
                if not profile:
                    profile = []
                boxplot_dict[col + "_outliers" + q_list].append(profile)

            df_box = pd.DataFrame(boxplot_dict)
            df_box["node"] = thicket.statsframe.dataframe.index.tolist()
            df_box.set_index("node", inplace=True)
            thicket.statsframe.dataframe = thicket.statsframe.dataframe.join(df_box)
    # columnar joined thicket object
    else:
        for idx, col in columns:
            boxplot_dict = {
                idx: {
                    col + "_q1" + q_list: [],
                    col + "_q2" + q_list: [],
                    col + "_q3" + q_list: [],
                    col + "_iqr" + q_list: [],
                    col + "_lowerfence" + q_list: [],
                    col + "_upperfence" + q_list: [],
                    col + "_outliers" + q_list: [],
                }
            }

            for node in pd.unique(thicket.dataframe.reset_index()["node"].tolist()):
                values = thicket.dataframe.loc[node][(idx, col)].tolist()

                q = np.quantile(values, quartiles)
                q1 = q[0]
                median = q[1]
                q3 = q[2]

                iqr = q3 - q1
                lower_fence = q1 - (1.5 * iqr)
                upper_fence = q3 + (1.5 * iqr)

                boxplot_dict[idx][col + "_q1" + q_list].append(q1)
                boxplot_dict[idx][col + "_q2" + q_list].append(median)
                boxplot_dict[idx][col + "_q3" + q_list].append(q3)
                boxplot_dict[idx][col + "_iqr" + q_list].append(iqr)
                boxplot_dict[idx][col + "_lowerfence" + q_list].append(lower_fence)
                boxplot_dict[idx][col + "_upperfence" + q_list].append(upper_fence)

                profile = []
                for i in range(0, len(values)):
                    if values[i] > upper_fence or values[i] < lower_fence:
                        profile.append(
                            thicket.dataframe[idx].loc[node].reset_index()["profile"][i]
                        )
                    else:
                        continue
                if not profile:
                    profile = []
                boxplot_dict[idx][col + "_outliers" + q_list].append(profile)

            create_multi_index = {
                (outerKey, innerKey): values
                for outerKey, innerDict in boxplot_dict.items()
                for innerKey, values in innerDict.items()
            }
            df_box = pd.DataFrame(create_multi_index)
            df_box["node"] = thicket.statsframe.dataframe.reset_index()["node"].tolist()
            df_box = df_box.set_index("node")
            thicket.statsframe.dataframe = thicket.statsframe.dataframe.join(df_box)

        # sort columns in index
        thicket.statsframe.dataframe = thicket.statsframe.dataframe.sort_index(axis=1)



# Copyright 2022 Lawrence Livermore National Security, LLC and other
# Thicket Project Developers. See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT

import numpy as np
import pandas as pd

from ..utils import verify_thicket_structures


def mean(thicket, columns=None):
    """Calculate the mean for each node in the performance data table.

    Designed to take in a thicket, and append one or more columns to the
    aggregated statistics table for the mean calculation for each node.

    Arguments:
        thicket (thicket): Thicket object
        columns (list): List of hardware/timing metrics to perform mean calculation on.
            Note, if using a columnar joined thicket a list of tuples must be passed in
            with the format (column index, column name).
    """
    if columns is None:
        raise ValueError(
            "To see a list of valid columns, please run Thicket.get_perf_columns()."
        )

    verify_thicket_structures(
        thicket.dataframe, index=["node", "profile"], columns=columns
    )

    # thicket object without columnar index
    if thicket.dataframe.columns.nlevels == 1:
        for column in columns:
            mean = []
            for node in pd.unique(thicket.dataframe.reset_index()["node"].tolist()):
                mean.append(np.mean(thicket.dataframe.loc[node][column]))
            thicket.statsframe.dataframe[column + "_mean"] = mean
    # columnar joined thicket object
    else:
        for idx, column in columns:
            mean = []
            for node in pd.unique(thicket.dataframe.reset_index()["node"].tolist()):
                mean.append(np.mean(thicket.dataframe.loc[node][(idx, column)]))
            thicket.statsframe.dataframe[(idx, column + "_mean")] = mean

        # sort columns in index
        thicket.statsframe.dataframe = thicket.statsframe.dataframe.sort_index(axis=1)



# Copyright 2022 Lawrence Livermore National Security, LLC and other
# Thicket Project Developers. See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT

import pandas as pd
import numpy as np
from scipy.stats import ttest_ind_from_stats
from scipy.stats import t

from thicket import Thicket


class StatsPreference(Thicket):
    """Determines a preference between two different thickets."""

    def __init__(
        self,
        thicket1,
        thicket2,
        columns=[],
        alpha=None,
        name_of_thicket1=None,
        name_of_thicket2=None,
        metadata_column=None,
    ):
        """Create a new statspreference object.

        Arguments:
            thicket1 (thicket): one thicket which will be compared to thicket2
            thicket2 (thicket): second thicket which will be compared to thicket1
            columns (list): list of 1 or more numeric columns to perform t-test on.
                Columns must be located in the performance data table.
            name_of_thicket1 (str, optional): descriptive name for thicket1
            name_of_thicket2 (str, optional): descriptive name for thicket2
            alpha (float, optional): configure the statistical significance threshold.
                Default is set to 0.05.
        """
        self.thicket1 = thicket1
        self.thicket2 = thicket2
        self.columns = columns
        self.metadata_column = metadata_column

        if alpha is None:
            self.alpha = 0.05
        else:
            self.alpha = alpha

        if name_of_thicket1 is None and name_of_thicket2 is None:
            self.combined_th = self.thicket1.columnar_join(
                self.thicket2, self.metadata_column
            )
            self.name_of_thicket1 = self.combined_th.dataframe.columns.levels[0][0]
            self.name_of_thicket2 = self.combined_th.dataframe.columns.levels[0][1]
        else:
            self.name_of_thicket1 = name_of_thicket1
            self.name_of_thicket2 = name_of_thicket2
            self.combined_th = self.thicket1.columnar_join(
                self.thicket2,
                self.metadata_column,
                self.name_of_thicket1,
                self.name_of_thicket2,
            )

        self.nobs1 = self.thicket1.dataframe.groupby(level=0).size()[0]
        self.nobs2 = self.thicket2.dataframe.groupby(level=0).size()[0]
        self.tvalue = t.ppf(q=1 - self.alpha / 2, df=self.nobs1 + self.nobs2 - 2)

    def preference(self, comparison_func, extra_cols=None):
        """Determine a preference between two different thickets.

        Arguments:
            comparison_func (callable function): function taking in two values,
                compared using a comparison operator such as < or >
            extra_cols (list, optional): list of 1 or more columns to be added to the
                initial instantiated column list

        Returns:
            (Thicket): a new thicket object with new column for preference
        """

        if extra_cols is not None:
            for new_cols in extra_cols:
                if new_cols not in self.columns:
                    self.columns.append(new_cols)

        self._ttest()

        for col in (
            self.combined_th.statsframe.dataframe["preference"]
            .filter(regex="tstatistic")
            .columns.tolist()
        ):
            for comparison_col in list(
                filter(
                    lambda x: col.strip("_")[0] in x,
                    self.combined_th.statsframe.dataframe[
                        self.name_of_thicket1
                    ].columns.tolist(),
                )
            ):
                preference = []
                for index in range(0, len(self.combined_th.statsframe.dataframe)):
                    if (
                        self.combined_th.statsframe.dataframe["preference"][col][index]
                        > self.tvalue
                        or self.combined_th.statsframe.dataframe["preference"][col][
                            index
                        ]
                        < -1 * self.tvalue
                    ):

                        val1 = self.combined_th.statsframe.dataframe[
                            self.name_of_thicket1
                        ][comparison_col][index]
                        val2 = self.combined_th.statsframe.dataframe[
                            self.name_of_thicket2
                        ][comparison_col][index]
                        preference.append(comparison_func(val1, val2))
                    else:
                        preference.append("No preference")

                self.combined_th.statsframe.dataframe[
                    ("preference", comparison_col + "_preferred")
                ] = preference

        self.combined_th.statsframe.dataframei = (
            self.combined_th.statsframe.dataframe.sort_index(axis=1)
        )

    def _ttest(self):
        """Calculate t-statistic.

        This method uses the number of observations, mean, and standard deviation
        from thicket1 and thicket2 to compute the t-statistic. The t-statistic is
        compared to the t-value to determine if there is a preference between thicket1
        and thicket2.
        """
        self._calc_mean()
        self._calc_std()

        col_means = (
            self.combined_th.statsframe.dataframe[self.name_of_thicket1]
            .filter(regex="mean")
            .columns.tolist()
        )
        col_stds = (
            self.combined_th.statsframe.dataframe[self.name_of_thicket1]
            .filter(regex="std")
            .columns.tolist()
        )

        for col_mean, col_std, col_name in zip(col_means, col_stds, self.columns):
            tStatisticList = []
            for i in range(0, len(self.combined_th.statsframe.dataframe)):
                tStatistic = ttest_ind_from_stats(
                    mean1=self.combined_th.statsframe.dataframe[self.name_of_thicket1][
                        col_mean
                    ][i],
                    std1=self.combined_th.statsframe.dataframe[self.name_of_thicket1][
                        col_std
                    ][i],
                    nobs1=self.nobs1,
                    mean2=self.combined_th.statsframe.dataframe[self.name_of_thicket2][
                        col_mean
                    ][i],
                    std2=self.combined_th.statsframe.dataframe[self.name_of_thicket2][
                        col_std
                    ][i],
                    nobs2=self.nobs2,
                    equal_var=False,
                )

                tStatisticList.append(tStatistic.statistic)
            self.combined_th.statsframe.dataframe[
                ("preference", "tvalue")
            ] = self.tvalue
            self.combined_th.statsframe.dataframe[
                ("preference", col_name + "_tstatistic")
            ] = tStatisticList

        self.combined_th.statsframe.dataframe = (
            self.combined_th.statsframe.dataframe.sort_index(axis=1)
        )

    def _calc_mean(self):
        """Calculates the mean value per node for thicket1 and thicket2."""
        for column in self.columns:
            mean_thicket1 = []
            mean_thicket2 = []
            for node in pd.unique(
                self.combined_th.dataframe.reset_index()["node"].tolist()
            ):
                mean1 = np.mean(
                    self.combined_th.dataframe.loc[node][
                        (self.name_of_thicket1, column)
                    ]
                )
                mean2 = np.mean(
                    self.combined_th.dataframe.loc[node][
                        (self.name_of_thicket2, column)
                    ]
                )
                mean_thicket1.append(mean1)
                mean_thicket2.append(mean2)

            self.combined_th.statsframe.dataframe[
                (self.name_of_thicket1, column + "_mean")
            ] = mean_thicket1
            self.combined_th.statsframe.dataframe[
                (self.name_of_thicket2, column + "_mean")
            ] = mean_thicket2

        self.combined_th.statsframe.dataframe = (
            self.combined_th.statsframe.dataframe.sort_index(axis=1)
        )

    def _calc_std(self):
        """Calculates the standard deviation per node for thicket1 and thicket2."""
        for column in self.columns:
            std_thicket1 = []
            std_thicket2 = []
            for node in pd.unique(
                self.combined_th.dataframe.reset_index()["node"].tolist()
            ):

                std1 = np.std(
                    self.combined_th.dataframe.loc[node][
                        (self.name_of_thicket1, column)
                    ]
                )
                std2 = np.std(
                    self.combined_th.dataframe.loc[node][
                        (self.name_of_thicket2, column)
                    ]
                )
                std_thicket1.append(std1)
                std_thicket2.append(std2)

            self.combined_th.statsframe.dataframe[
                (self.name_of_thicket1, column + "_std")
            ] = std_thicket1
            self.combined_th.statsframe.dataframe[
                (self.name_of_thicket2, column + "_std")
            ] = std_thicket2

        self.combined_th.statsframe.dataframe = (
            self.combined_th.statsframe.dataframe.sort_index(axis=1)
        )



# Copyright 2022 Lawrence Livermore National Security, LLC and other
# Thicket Project Developers. See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT

import pandas as pd
import seaborn as sns
import hatchet as ht

from ..utils import verify_thicket_structures


def display_histogram(thicket, node=None, column=None, **kwargs):
    """Display a histogram for a user passed node and column. Node and column must come
    from the performance data table.

    Arguments:
        thicket (thicket): Thicket object
        node (node): Node object
        column (str): Column from performance data table. Note, if using a
            columnar joined thicket a tuple must be passed in with the format
            (column index, column name).

    Returns:
        (matplotlib Axes): Object for managing histogram plot.
    """
    if column is None or node is None:
        raise ValueError(
            "To see a list of valid columns, run Thicket.get_perf_columns()."
        )

    if not isinstance(node, ht.node.Node):
        raise ValueError(
            "Value passed to node argument must be of type hatchet.node.Node."
        )

    verify_thicket_structures(
        thicket.dataframe, index=["node", "profile"], columns=[column]
    )

    # thicket object without columnar index
    if thicket.dataframe.columns.nlevels == 1:
        df = pd.melt(
            thicket.dataframe.reset_index(),
            id_vars="node",
            value_vars=column,
            value_name=" ",
        )

        filtered_df = df[df["node"] == node]

        ax = sns.displot(filtered_df, x=" ", kind="hist", **kwargs)

        return ax
    # columnar joined thicket object
    else:
        col_idx, column_value = column[0], column[1]
        df_subset = thicket.dataframe[col_idx]

        df = pd.melt(
            df_subset.reset_index(),
            id_vars="node",
            value_vars=column_value,
            value_name=" ",
        )

        filtered_df = df[df["node"] == node]

        ax = sns.displot(filtered_df, x=" ", kind="hist", **kwargs)

        return ax



# Copyright 2022 Lawrence Livermore National Security, LLC and other
# Thicket Project Developers. See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT

import pandas as pd
import seaborn as sns
import hatchet as ht

from ..utils import verify_thicket_structures


def display_boxplot(thicket, nodes=[], columns=[], **kwargs):
    """Display a boxplot for each user passed node(s) and column(s). The passed nodes
    and columns must be from the performance data table.

    Designed to take in a thicket, and display a plot with one or more boxplots
    depending on the number of nodes and columns passed.

    Arguments:
        thicket (thicket): Thicket object
        nodes (list): List of nodes to view on the x-axis
        column (list): List of hardware/timing metrics to view on the y-axis. Note, if
            using a columnar joined thicket a list of tuples must be passed in with the
            format (column index, column name).

    Returns:
        (matplotlib Axes): Object for managing boxplot.
    """
    for node in nodes:
        if not isinstance(node, ht.node.Node):
            raise ValueError(
                "Value(s) passed to node argument must be of type hatchet.node.Node."
            )

    verify_thicket_structures(
        thicket.dataframe, index=["node", "profile"], columns=columns
    )

    # thicket object without columnar index
    if thicket.dataframe.columns.nlevels == 1:
        df = pd.melt(
            thicket.dataframe.reset_index(),
            id_vars=["node", "name"],
            value_vars=columns,
            var_name="Performance counter",
            value_name=" ",
        )

        position = []
        for node in nodes:
            idx = df.index[df["node"] == node]
            for pos in idx:
                position.append(pos)

        # rename columns such that the x-axis label is "node" and not "name", tick marks
        # will be node names
        filtered_df = df.loc[position].rename(
            columns={"node": "hatchet node", "name": "node"}
        )

        if len(columns) > 1:
            return sns.boxplot(
                data=filtered_df, x="node", y=" ", hue="Performance counter", **kwargs
            )
        else:
            return sns.boxplot(data=filtered_df, x="node", y=" ", **kwargs)
    # columnar joined thicket object
    else:
        initial_idx, initial_col = columns[0][0], columns[0][1]
        cols = [initial_col]
        for i in range(1, len(columns)):
            if initial_idx != columns[i][0]:
                raise ValueError(
                    "Columns specified as tuples must have the same column index (first element)"
                )
            else:
                cols.append(columns[i][1])

        df_subset = thicket.dataframe[initial_idx].reset_index()
        df_subset["name"] = thicket.dataframe["name"].tolist()

        df = pd.melt(
            df_subset,
            id_vars=["node", "name"],
            value_vars=cols,
            var_name="Performance counter",
            value_name=" ",
        )

        position = []
        for node in nodes:
            idx = df.index[df["node"] == node]
            for pos in idx:
                position.append(pos)

        # rename columns such that the x-axis label is "node" and not "name", tick marks
        # will be node names
        filtered_df = df.loc[position].rename(
            columns={"node": "hatchet node", "name": "node"}
        )

        if len(columns) > 1:
            return sns.boxplot(
                data=filtered_df, x="node", y=" ", hue="Performance counter", **kwargs
            )
        else:
            return sns.boxplot(data=filtered_df, x="node", y=" ", **kwargs)



# Copyright 2022 Lawrence Livermore National Security, LLC and other
# Thicket Project Developers. See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT

import base64
from io import BytesIO
from statistics import mean

import extrap.entities as xent
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from extrap.entities.experiment import (
    Experiment,
)  # For some reason it errors if "Experiment" is not explicitly imported
from extrap.fileio import io_helper
from extrap.modelers.model_generator import ModelGenerator


MODEL_TAG = "_extrap-model"


class ModelWrapper:
    """Wrapper for an Extra-P model.

    Provides more convenient functions for evaluating the model at given data points,
    for writing out a string representation of the model, and for displaying (plotting)
    the model.
    """

    def __init__(self, mdl, param_name):
        self.mdl = mdl
        self.param_name = param_name  # Needed for plotting / displaying the model

    def __str__(self):
        """Display self as a function"""
        return str(self.mdl.hypothesis.function)

    def eval(self, val):
        """Evaluate function (self) at val. f(val) = result"""
        return self.mdl.hypothesis.function.evaluate(val)

    def display(self, RSS):
        """Display function

        Arguments:
            RSS (bool): whether to display Extra-P RSS on the plot
        """
        # Sort based on x values
        measures_sorted = sorted(self.mdl.measurements, key=lambda x: x.coordinate[0])

        # Scatter plot
        params = [ms.coordinate[0] for ms in measures_sorted]  # X values
        measures = [ms.value(True) for ms in measures_sorted]  # Y values

        # Line plot

        # X value plotting range. Dynamic based off what the largest/smallest values are
        x_vals = np.arange(
            params[0], 1.5 * params[-1], (params[-1] - params[0]) / 100.0
        )

        # Y values
        y_vals = [self.mdl.hypothesis.function.evaluate(x) for x in x_vals]

        plt.ioff()
        fig, ax = plt.subplots()

        # Plot line
        ax.plot(x_vals, y_vals, label=self.mdl.hypothesis.function)

        # Plot scatter
        ax.plot(params, measures, "ro", label=self.mdl.callpath)

        ax.set_xlabel(self.param_name)
        ax.set_ylabel(self.mdl.metric)
        if RSS:
            ax.text(
                x_vals[0],
                max(y_vals + measures),
                "RSS = " + self.mdl.hypothesis.RSS,
            )
        ax.legend(loc=1)

        return fig, ax


class Modeling:
    """Produce models for all the metrics across the given graphframes."""

    def __init__(self, tht, param_name, params=None, chosen_metrics=None):
        """Create a new model object.

        Adds a model column for each metric for each common frame across all the
        graphframes.

        The given list of params contains the parameters to build the models.  For
        example, MPI ranks, input sizes, and so on.

        Arguments:
            tht (Thicket): thicket object
            param_name (str): arbitrary if 'params' is being provided, otherwise name of
                the metadata column from which 'params' will be extracted
            params (list): parameters list, domain for the model
            chosen_metrics (list): metrics to be evaluated in the model, range for the
                model
        """
        self.tht = tht
        self.param_name = param_name

        # Assign param
        # Get params from metadata table
        if not params:
            self.params = self.tht.metadata[param_name].tolist()
        # params must be provided by the user
        else:
            if not isinstance(params, dict):
                raise TypeError("'params' must be provided as a dict")
            elif len(params) != len(self.tht.profile):
                raise ValueError(
                    "length of params must equal amount of profiles "
                    + len(params)
                    + "!= "
                    + len(self.tht.profile)
                )
            profile_mapping_flipped = {
                v[0]: k for k, v in self.tht.profile_mapping.items()
            }
            for file_name, value in params.items():
                self.tht.metadata.at[
                    profile_mapping_flipped[file_name], param_name
                ] = value
            self.params = tht.metadata[param_name].tolist()
        if not chosen_metrics:
            self.chosen_metrics = self.tht.exc_metrics + self.tht.inc_metrics
        else:
            self.chosen_metrics = chosen_metrics

    def to_html(self, RSS=False):
        def model_to_img_html(model_obj):
            fig, ax = model_obj.display(RSS)
            figfile = BytesIO()
            fig.savefig(figfile, format="jpg", transparent=False)
            figfile.seek(0)
            figdata_jpg = base64.b64encode(figfile.getvalue()).decode()
            imgstr = '<img src="data:image/jpg;base64,{}" />'.format(figdata_jpg)
            plt.close(fig)
            return imgstr

        frm_dict = {met + MODEL_TAG: model_to_img_html for met in self.chosen_metrics}

        # Subset of the aggregated statistics table with only the Extra-P columns selected
        return self.tht.statsframe.dataframe[
            [met + MODEL_TAG for met in self.chosen_metrics]
        ].to_html(escape=False, formatters=frm_dict)

    def _add_extrap_statistics(self, node, metric):
        """Insert the Extra-P hypothesis function statistics into the aggregated
            statistics table. Has to be called after "produce_models".

        Arguments:
            node (hatchet.Node): The node for which statistics should be calculated
            metric (str): The metric for which statistics should be calculated
        """
        hypothesis_fn = self.tht.statsframe.dataframe.at[
            node, metric + MODEL_TAG
        ].mdl.hypothesis

        self.tht.statsframe.dataframe.at[
            node, metric + "_RSS" + MODEL_TAG
        ] = hypothesis_fn.RSS
        self.tht.statsframe.dataframe.at[
            node, metric + "_rRSS" + MODEL_TAG
        ] = hypothesis_fn.rRSS
        self.tht.statsframe.dataframe.at[
            node, metric + "_SMAPE" + MODEL_TAG
        ] = hypothesis_fn.SMAPE
        self.tht.statsframe.dataframe.at[
            node, metric + "_AR2" + MODEL_TAG
        ] = hypothesis_fn.AR2
        self.tht.statsframe.dataframe.at[
            node, metric + "_RE" + MODEL_TAG
        ] = hypothesis_fn.RE

    def produce_models(self, agg_func=mean, add_stats=True):
        """Produces an Extra-P model. Models are generated by calling Extra-P's
            ModelGenerator.

        Arguments:
            agg_func (function): aggregation function to apply to multi-dimensional
                measurement values. Extra-P v4.0.4 applies mean by default so that is
                set here for clarity.
            add_stats (bool): Option to add hypothesis function statistics to the
                aggregated statistics table
        """
        # Setup domain values one time. Have to match ordering with range
        # values (i.e. performance data table profile ordering)
        param_coords = []  # default coordinates for all profiles

        # Mapping from metadata profiles to the parameter
        meta_param_mapping = self.tht.metadata[self.param_name].to_dict()

        # Flipped version of mapping dictionary
        meta_param_mapping_flipped = dict(
            [(value, key) for key, value in meta_param_mapping.items()]
        )

        # Ordering of profiles in the performance data table
        ensemble_profile_ordering = list(self.tht.dataframe.index.unique(level=1))

        # Append coordinates in order
        for profile in ensemble_profile_ordering:
            param_coords.append(
                xent.coordinate.Coordinate(float(meta_param_mapping[profile]))
            )

        # Iterate over nodes (outer index)
        for node, single_node_df in self.tht.dataframe.groupby(level=0):
            # Start experiment
            exp = Experiment()
            exp.add_parameter(xent.parameter.Parameter(self.param_name))

            # For all chosen metrics
            for met in self.chosen_metrics:
                measures = []
                # Iterate over profiles (secondary index)
                for profile, single_prof_df in single_node_df.groupby(level=1):
                    measures.append(single_prof_df[met].tolist())
                # Apply aggregation function to measurements
                for i in range(len(measures)):
                    # Only apply aggregation function if more that one value
                    if isinstance(measures[i], list):
                        measures[i] = agg_func(measures[i])
                # Add coordinates (points at which measurements were taken) for the
                # default case
                if len(measures) == len(param_coords):
                    exp.coordinates.extend(param_coords)
                # Handle case where profile(s) do not contain a measurement for the
                # current node
                elif len(measures) < len(param_coords):
                    param_coords_subset = []
                    df_index = self.tht.dataframe.loc[node, met].index
                    for coord in param_coords:
                        profile = meta_param_mapping_flipped[coord._values[0]]
                        if df_index.isin([profile], level=0).any():
                            param_coords_subset.append(coord)
                        else:
                            print(
                                "(Coordinate removed) Measurement at ("
                                + profile
                                + "): "
                                + meta_param_mapping[profile]
                                + "DNE for node "
                                + node
                            )
                    exp.coordinates.extend(param_coords_subset)

                # Create metric object
                metric_obj = xent.metric.Metric(met)
                exp.add_metric(xent.metric.Metric(met))
                # Create callpath object and call tree
                cpath = xent.callpath.Callpath(node.frame["name"])
                exp.add_callpath(cpath)
                exp.call_tree = io_helper.create_call_tree(exp.callpaths)
                # Add measurement objects to experiment
                for coord, measurement in zip(exp.coordinates, measures):
                    measurement_obj = xent.measurement.Measurement(
                        coord, cpath, metric_obj, measurement
                    )
                    exp.add_measurement(measurement_obj)
                # Sanity check
                io_helper.validate_experiment(exp)
                # Generate model
                model_gen = ModelGenerator(exp)
                model_gen.model_all()
                mkey = (cpath, metric_obj)
                self.tht.statsframe.dataframe.at[node, met + MODEL_TAG] = ModelWrapper(
                    model_gen.models[mkey], self.param_name
                )
                # Add statistics to aggregated statistics table
                if add_stats:
                    self._add_extrap_statistics(node, met)

    def _componentize_function(model_object):
        """Componentize one Extra-P modeling object into a dictionary of its parts

        Arguments:
            model_object (ModelWrapper): Thicket ModelWrapper Extra-P modeling object

        Returns:
            (dict): dictionary of the ModelWrapper's hypothesis function parts
        """
        # Dictionary of variables mapped to coefficients
        term_dict = {}
        # Model object hypothesis function
        fnc = model_object.mdl.hypothesis.function
        # Constant "c" column
        term_dict["c"] = fnc.constant_coefficient

        # Terms of form "coefficient * variables"
        for term in fnc.compound_terms:
            # Join variables of the same term together
            variable_column = " * ".join(t.to_string() for t in term.simple_terms)

            term_dict[variable_column] = term.coefficient

        return term_dict

    def componentize_statsframe(self, columns=None):
        """Componentize multiple Extra-P modeling objects in the aggregated statistics
        table

        Arguments:
            column (list): list of column names in the aggregated statistics table to
                componentize. Values must be of type 'thicket.model_extrap.ModelWrapper'.
        """
        # Use all Extra-P columns
        if columns is None:
            columns = [
                col
                for col in self.tht.statsframe.dataframe
                if isinstance(self.tht.statsframe.dataframe[col][0], ModelWrapper)
            ]

        # Error checking
        for c in columns:
            if c not in self.tht.statsframe.dataframe.columns:
                raise ValueError(
                    "column " + c + " is not in the aggregated statistics table."
                )
            elif not isinstance(self.tht.statsframe.dataframe[c][0], ModelWrapper):
                raise TypeError(
                    "column "
                    + c
                    + " is not the right type (thicket.model_extrap.ModelWrapper)."
                )

        # Process each column
        all_dfs = []
        for col in columns:
            # Get list of components for this column
            components = [
                Modeling._componentize_function(model_obj)
                for model_obj in self.tht.statsframe.dataframe[col]
            ]

            # Component dataframe
            comp_df = pd.DataFrame(
                data=components, index=self.tht.statsframe.dataframe.index
            )

            # Add column name as index level
            comp_df.columns = pd.MultiIndex.from_product(
                [[col], comp_df.columns.to_list()]
            )
            all_dfs.append(comp_df)

        # Concatenate dataframes horizontally
        all_dfs.insert(0, self.tht.statsframe.dataframe)
        self.tht.statsframe.dataframe = pd.concat(all_dfs, axis=1)



# Copyright 2022 Lawrence Livermore National Security, LLC and other
# Thicket Project Developers. See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT

import pandas as pd


def _are_synced(gh, df):
    """Check if node objects are equal in graph and dataframe id(graph_node) ==
    id(df_node).
    """
    for graph_node in gh.traverse():
        node_present = False
        for df_node in df.index.get_level_values("node"):
            if id(df_node) == id(graph_node):
                node_present = True
                continue
        if not node_present:
            return False
    return True


def _missing_nodes_to_list(a_df, b_df):
    """Get a list of node differences between two dataframes. Mainly used for "tree"
    function.

    Arguments:
        a_df (Dataframe): First pandas Dataframe
        b_df (Dataframe): Second pandas Dataframe

    Returns:
        (list): List of numbers in range (0, 1, 2). "0" means node is in both, "1" is
            only in "a", "2" is only in "b"
    """
    missing_nodes = []
    a_list = list(map(hash, list(a_df.index.get_level_values("node"))))
    b_list = list(map(hash, list(b_df.index.get_level_values("node"))))
    # Basic cases
    while a_list and b_list:
        a = a_list.pop(0)
        b = b_list.pop(0)
        while a > b and b_list:
            missing_nodes.append(2)
            b = b_list.pop(0)
        while b > a and a_list:
            missing_nodes.append(1)
            a = a_list.pop(0)
        if a == b:
            missing_nodes.append(0)
            continue
        elif a > b:  # Case where last two nodes and "a" is missing "b" then opposite
            missing_nodes.append(2)
            missing_nodes.append(1)
        elif b > a:  # Case where last two nodes and "b" is missing "a" then opposite
            missing_nodes.append(1)
            missing_nodes.append(2)
    while a_list:  # In case "a" has a lot more nodes than "b"
        missing_nodes.append(1)
        a = a_list.pop(0)
    while b_list:  # In case "b" has a lot more nodes than "a"
        missing_nodes.append(2)
        b = b_list.pop(0)
    return missing_nodes


def _new_statsframe_df(df, multiindex=False):
    """Generate new aggregated statistics table from a dataframe. This is most commonly
    needed when changes are made to the performance data table's index.

    Arguments:
        df (DataFrame): Input dataframe to generate the aggregated statistics table from
        multiindex (Bool, optional): Option to setup MultiIndex column structure. This
            is standard to do if performance data table is MultiIndex.

    Returns:
        (DataFrame): new aggregated statistics table
    """
    nodes = list(set(df.reset_index()["node"]))  # List of nodes
    names = [node.frame["name"] for node in nodes]  # List of names

    # Create new dataframe with "node" index and "name" data by default.
    new_df = pd.DataFrame(
        data={"node": nodes, "name": names},
    ).set_index("node")

    # Create MultiIndex structure if necessary.
    if multiindex:
        new_df.columns = pd.MultiIndex.from_tuples([("", "name")])

    return new_df


def _print_graph(graph):
    """Print the nodes in a hatchet graph"""
    i = 0
    for node in graph.traverse():
        print(f"{node} ({hash(node)}) ({id(node)})")
        i += 1
    return i


def _resolve_missing_indicies(th_list):
    """Resolve indices if at least 1 profile has an index that another doesn't.

    If at least one profile has an index that another doesn't, then issues will arise
    when unifying. Need to add this index to other thickets.

    Note that the value to use for the new index is set to '0' for ease-of-use, but
    something like 'NaN' may arguably provide more clarity.
    """
    # Create a set of all index possibilities
    idx_set = set({})
    for th in th_list:
        idx_set.update(th.dataframe.index.names)

    # Apply missing indicies to thickets
    for th in th_list:
        for idx in idx_set:
            if idx not in th.dataframe.index.names:
                print(f"Resolving '{idx}' in thicket: ({id(th)})")
                th.dataframe[idx] = 0
                th.dataframe.set_index(idx, append=True, inplace=True)


def _sync_nodes(gh, df):
    """Set the node objects to be equal in both the graph and the dataframe.

    id(graph_node) == id(df_node) after this function for nodes with equivalent hatchet
    nid's.
    """
    index_names = df.index.names
    df.reset_index(inplace=True)
    for graph_node in gh.traverse():
        df["node"] = df["node"].apply(
            lambda df_node: graph_node
            if (hash(graph_node) == hash(df_node))
            else df_node
        )
    df.set_index(index_names, inplace=True)


def _sync_nodes_frame(gh, df):
    """Update dataframe node objects hnid's based off their positioning relative to a
    given graph.

    id(graph_node) == id(df_node) after this function for nodes with equivalent hatchet
    nid's.

    TODO: This function may be superior to _sync_nodes and may be able to replace it.
    Need to investigate.
    """
    assert df.index.nlevels == 2  # For num_profiles assumption

    # TODO: Graph function to list conversion: move to Hatchet?
    gh_node_list = []
    for gh_node in gh.traverse():
        gh_node_list.append(gh_node)

    num_profiles = len(df.groupby(level=1))
    index_names = df.index.names
    df.reset_index(inplace=True)
    df_node_list = df["node"][::num_profiles].to_list()

    # Sequentially walk through graph and dataframe and modify dataframe hnid's based off graph equivalent
    i = 0
    j = 0
    while i < len(gh_node_list) and j < len(df_node_list):
        if gh_node_list[i].frame == df_node_list[j].frame:
            df_node_list[j] = gh_node_list[i]
            j += 1
        else:
            i += 1

    # Extend list to match multi-index dataframe structure
    df_list_full = []
    for node in df_node_list:
        temp = []
        for idx in range(num_profiles):
            temp.append(node)
        df_list_full.extend(temp)
    # Update nodes in the dataframe
    df["node"] = df_list_full

    df.set_index(index_names, inplace=True)


def _get_perf_columns(df):
    """Get list of performance dataframe columns that are numeric.

    Numeric columns can be used with thicket's statistical functions.

    Arguments:
        df (DataFrame): thicket dataframe object

    Returns:
        numeric_columns (list): list of numeric columns
    """
    numeric_types = ["int32", "int64", "float32", "float64"]

    numeric_columns = df.select_dtypes(include=numeric_types).columns.tolist()

    # thicket object without columnar index
    if df.columns.nlevels == 1:
        if "nid" in numeric_columns:
            numeric_columns.remove("nid")

        return numeric_columns
    # columnar joined thicket object
    else:
        return [x for x in numeric_columns if "nid" not in x]



# Copyright 2022 Lawrence Livermore National Security, LLC and other
# Thicket Project Developers. See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT

from scipy import stats

from ..utils import verify_thicket_structures


def correlation_nodewise(thicket, column1=None, column2=None, correlation="pearson"):
    """Calculate the nodewise correlation for each node in the performance data table.

    Designed to take in a thicket, and append one or more columns to the aggregated
    statistics table for the nodewise correlation calculation for each node.

    Arguments:
        thicket (thicket): Thicket object
        column1 (str): First comparison column. Note, if using a columnar joined thicket
            a tuple must be passed in with the format (column index, column name).
        column2 (str): Second comparison column. Note, if using a columnar joined
            thicket a tuple must be passed in with the format
            (column index, column name).
        correlation (str): correlation test to perform -- pearson (default), spearman,
            and kendall.
    """
    if column1 is None or column2 is None:
        raise ValueError(
            "To see a list of valid columns, please run Thicket.get_perf_columns()."
        )

    if "profile" in thicket.dataframe.index.names:
        verify_thicket_structures(
            thicket.dataframe,
            index=["node", "profile"],
            columns=[column1, column2],
        )
    else:
        verify_thicket_structures(
            thicket.dataframe,
            index=["node", "thicket"],
            columns=[column1, column2],
        )

    # thicket object without columnar index
    if thicket.dataframe.columns.nlevels == 1:
        correlated = []
        for node in thicket.statsframe.dataframe.index.tolist():
            if correlation == "pearson":
                correlated.append(
                    stats.pearsonr(
                        thicket.dataframe.loc[node][column1],
                        thicket.dataframe.loc[node][column2],
                    )[0]
                )
            elif correlation == "spearman":
                correlated.append(
                    stats.spearmanr(
                        thicket.dataframe.loc[node][column1],
                        thicket.dataframe.loc[node][column2],
                    )[0]
                )
            elif correlation == "kendall":
                correlated.append(
                    stats.kendalltau(
                        thicket.dataframe.loc[node][column1],
                        thicket.dataframe.loc[node][column2],
                    )[0]
                )
            else:
                raise ValueError("Invalid correlation")
        thicket.statsframe.dataframe[
            column1 + "_vs_" + column2 + " " + correlation
        ] = correlated
    # columnar joined thicket object
    else:
        correlated = []
        for node in thicket.statsframe.dataframe.index.tolist():
            if correlation == "pearson":
                correlated.append(
                    stats.pearsonr(
                        thicket.dataframe.loc[node][column1],
                        thicket.dataframe.loc[node][column2],
                    )[0]
                )
            elif correlation == "spearman":
                correlated.append(
                    stats.spearmanr(
                        thicket.dataframe.loc[node][column1],
                        thicket.dataframe.loc[node][column2],
                    )[0]
                )
            elif correlation == "kendall":
                correlated.append(
                    stats.kendalltau(
                        thicket.dataframe.loc[node][column1],
                        thicket.dataframe.loc[node][column2],
                    )[0]
                )
            else:
                raise ValueError(
                    "Invalid correlation, options are pearson, spearman, and kendall."
                )
        if column1[0] != column2[0]:
            thicket.statsframe.dataframe[
                (
                    "Union statistics",
                    column1[1] + "_vs_" + column2[1] + " " + correlation,
                )
            ] = correlated
        else:
            column_idx = column1[0]
            thicket.statsframe.dataframe[
                (column_idx, column1[1] + "_vs_" + column2[1] + " " + correlation)
            ] = correlated

        # sort columns in index
        thicket.statsframe.dataframe = thicket.statsframe.dataframe.sort_index(axis=1)



# Copyright 2022 Lawrence Livermore National Security, LLC and other
# Thicket Project Developers. See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT

import pandas as pd
from scipy import stats

from ..utils import verify_thicket_structures


def check_normality(thicket, columns=None):
    """Determine if the data is normal or non-normal for each node in the performance
    data table.

    Designed to take in a thicket, and append one or more columns to the aggregated
    statistics table. A true boolean value will be appended if the data is normal and a
    false boolean value will be appended if the data is non-normal.

    For this test, the more data the better. Preferably you would want to have 20 data
    points (20 files) in a dataset to have an accurate result.

    Arguments:
        thicket (thicket): Thicket object
        columns (list): List of hardware/timing metrics to perform normality test on.
            Note, if using a columnar joined thicket a list of tuples must be passed in
            with the format (column index, column name).
    """
    if columns is None:
        raise ValueError(
            "To see a list of valid columns, please run Thicket.get_perf_columns()."
        )

    verify_thicket_structures(
        thicket.dataframe, index=["node", "profile"], columns=columns
    )

    # thicket object without columnar index
    if thicket.dataframe.columns.nlevels == 1:
        for column in columns:
            normality = []
            for node in pd.unique(thicket.dataframe.reset_index()["node"].tolist()):
                pvalue = stats.shapiro(thicket.dataframe.loc[node][column])[1]

                if pvalue < 0.05:
                    normality.append("False")
                elif pvalue > 0.05:
                    normality.append("True")
                else:
                    normality.append(pd.NA)

            thicket.statsframe.dataframe[column + "_normality"] = normality
    # columnar joined thicket object
    else:
        for idx, column in columns:
            normality = []
            for node in pd.unique(thicket.dataframe.reset_index()["node"].tolist()):
                pvalue = stats.shapiro(thicket.dataframe.loc[node][(idx, column)])[1]

                if pvalue < 0.05:
                    normality.append("False")
                elif pvalue > 0.05:
                    normality.append("True")
                else:
                    normality.append(pd.NA)

            thicket.statsframe.dataframe[(idx, column + "_normality")] = normality

        # sort columns in index
        thicket.statsframe.dataframe = thicket.statsframe.dataframe.sort_index(axis=1)



from os import path
from os.path import dirname

from IPython.core.magic import Magics, magics_class, line_magic
from hatchet.external import Roundtrip as RT


def _thicket_to_json(data):
    return data.to_json()


def _df_to_json(data):
    return data.to_json(orient="records")


def _basic_to_json(data):
    import json

    return json.dumps(data)


vis_dir = dirname(path.abspath(__file__))


@magics_class
class EnsembleVis(Magics):
    def __init__(self, shell):
        super(EnsembleVis, self).__init__(shell)
        self.vis_dist = path.join(vis_dir, "static")

    @line_magic
    def metadata_vis(self, line):
        args = line.split(" ")
        RT.load_webpack(path.join(self.vis_dist, "pcp_bundle.html"), cache=False)
        RT.var_to_js(
            args[0], "thicket_ensemble", watch=False, to_js_converter=_thicket_to_json
        )

        if len(args) > 1:
            RT.var_to_js(
                args[1], "metadata_dims", watch=False, to_js_converter=_basic_to_json
            )

        if len(args) > 2:
            RT.var_to_js(
                args[2], "focus_node", watch=False, to_js_converter=_basic_to_json
            )

        RT.initialize()

    @line_magic
    def topdown_analysis(self, line):
        args = line.split(" ")
        RT.load_webpack(path.join(self.vis_dist, "topdown_bundle.html"), cache=False)
        RT.var_to_js(
            args[0], "topdown_data", watch=False, to_js_converter=_thicket_to_json
        )

        RT.initialize()


def load_ipython_extension(ipython):
    ipython.register_magics(EnsembleVis)



# Copyright 2022 Lawrence Livermore National Security, LLC and other
# Thicket Project Developers. See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT

from setuptools import setup
from codecs import open
from os import path


def readme():
    here = path.abspath(path.dirname(__file__))
    with open(path.join(here, "README.md"), encoding="utf-8") as f:
        return f.read()


# Get the version in a safe way which does not reference thicket `__init__` file
# per python docs: # https://packaging.python.org/guides/single-sourcing-package-version/
version = {}
with open("./thicket/version.py") as fp:
    exec(fp.read(), version)


setup(
    name="llnl-thicket",
    version=version["__version__"],
    license="MIT",
    description="Toolkit for exploratory data analysis of ensemble performance data",
    long_description=readme(),
    long_description_content_type="text/markdown",
    keywords="",
    project_urls={
        "Source Code": "https://github.com/LLNL/thicket",
        "Documentation": "https://thicket.readthedocs.io/",
    },
    python_requires=">=3.6.1",
    packages=[
        "thicket",
        "thicket.stats",
        "thicket.vis",
    ],
    include_package_data=True,
    install_requires=[
        "scipy",
        "seaborn",
        "pydot",
        "matplotlib",
        "numpy",
        "pandas >= 1.1",
        "llnl-hatchet",
    ],
    extras_require={"extrap": ["extrap"]},
)



# Copyright 2022 Lawrence Livermore National Security, LLC and other
# Thicket Project Developers. See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT

import numpy as np
import pandas as pd

from ..utils import verify_thicket_structures


def percentiles(thicket, columns=None):
    """Calculate the q-th percentile for each node in the performance data table.

    Designed to take in a thicket, and append one or more columns to the aggregated
    statistics table for the q-th percentile calculation for each node.

    The 25th percentile is the lower quartile, and is the value at which 25% of the
    answers lie below that value.

    The 50th percentile, is the median and half of the values lie below the median and
    half lie above the median.

    The 75th percentile is the upper quartile, and is the value at which 25% of the
    answers lie above that value and 75% of the answers lie below that value.

    Arguments:
        thicket (thicket): Thicket object
        columns (list): List of hardware/timing metrics to perform percentile
            calculation on. Note if using a columnar joined thicket a list of tuples
            must be passed in with the format (column index, column name).
    """
    if columns is None:
        raise ValueError(
            "To see a list of valid columns, please run Thicket.get_perf_columns()."
        )

    verify_thicket_structures(
        thicket.dataframe, index=["node", "profile"], columns=columns
    )

    # thicket object without columnar index
    if thicket.dataframe.columns.nlevels == 1:
        for column in columns:
            percentiles = []
            for node in pd.unique(thicket.dataframe.reset_index()["node"].tolist()):
                percentiles.append(
                    np.percentile(thicket.dataframe.loc[node][column], [25, 50, 75])
                )
            thicket.statsframe.dataframe[column + "_percentiles"] = percentiles
    # columnar joined thicket object
    else:
        for idx, column in columns:
            percentiles = []
            for node in pd.unique(thicket.dataframe.reset_index()["node"].tolist()):
                percentiles.append(
                    np.percentile(
                        thicket.dataframe.loc[node][(idx, column)], [25, 50, 75]
                    )
                )
            thicket.statsframe.dataframe[(idx, column + "_percentiles")] = percentiles

        # sort columns in index
        thicket.statsframe.dataframe = thicket.statsframe.dataframe.sort_index(axis=1)



# Copyright 2022 Lawrence Livermore National Security, LLC and other
# Thicket Project Developers. See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT

import copy
import os
import json
import warnings
from collections import OrderedDict

import pandas as pd
import numpy as np
from hatchet import GraphFrame
from hatchet.query import AbstractQuery

import thicket.helpers as helpers
from .utils import verify_sorted_profile
from .utils import verify_thicket_structures


class Thicket(GraphFrame):
    """Ensemble of profiles, includes a graph and three dataframes, performance data,
    metadata, and aggregated statistics.
    """

    def __init__(
        self,
        graph,
        dataframe,
        exc_metrics=None,
        inc_metrics=None,
        default_metric="time",
        metadata={},
        performance_cols=None,
        profile=None,
        profile_mapping=None,
        statsframe=None,
    ):
        """Create a new thicket from a graph and a dataframe.

        Arguments:
            graph (Graph): graph of nodes in this thicket
            dataframe (DataFrame): pandas DataFrame indexed by Nodes from the graph, and
                potentially other indexes
            exc_metrics (list): list of names of exclusive metrics in the dataframe
            inc_metrics (list): list of names of inclusive metrics in the dataframe
            default_metric (str): primary metric
            metadata (DataFrame): pandas DataFrame indexed by profile hashes, contains
                profile metadata
            performance_cols (list): list of numeric columns within the performance
                dataframe
            profile (list): list of hashed profile strings
            profile_mapping (dict): mapping of hashed profile strings to original strings
            statsframe (DataFrame): pandas DataFrame indexed by Nodes from the graph
        """
        super().__init__(
            graph, dataframe, exc_metrics, inc_metrics, default_metric, metadata
        )
        self.profile = profile
        self.profile_mapping = profile_mapping
        if statsframe is None:
            self.statsframe = GraphFrame(
                graph=self.graph,
                dataframe=helpers._new_statsframe_df(dataframe),
            )
        else:
            self.statsframe = statsframe

        self.performance_cols = helpers._get_perf_columns(self.dataframe)

    def __eq__(self, other):
        """Compare two thicket objects.

        Arguments:
            other (Thicket): Thicket object to compare to

        Returns:
            (bool): True if equal, False otherwise
        """
        assert isinstance(other, Thicket)
        return (
            self.graph == other.graph
            and self.dataframe.equals(other.dataframe)
            and self.exc_metrics == other.exc_metrics
            and self.inc_metrics == other.inc_metrics
            and self.default_metric == other.default_metric
            and self.metadata.equals(other.metadata)
            and self.performance_cols == other.performance_cols
            and self.profile == other.profile
            and self.profile_mapping == other.profile_mapping
            and self.statsframe.graph == other.statsframe.graph
            and self.statsframe.dataframe.equals(other.statsframe.dataframe)
        )

    def __str__(self):
        s = (
            "graph: "
            + helpers._print_graph(self.graph)
            + "\ndataframe:\n"
            + self.dataframe
            + "\nexc_metrics: "
            + self.exc_metrics
            + "\ninc_metrics: "
            + self.inc_metrics
            + "\ndefault_metric: "
            + self.default_metric
            + "\nmetadata:\n"
            + self.metadata
            + "\nperformance_cols:\n"
            + self.performance_cols
            + "\nprofile: "
            + self.profile
            + "\nprofile_mapping: "
            + self.profile_mapping
            + "\nstatsframe:\n"
            + helpers._print_graph(self.statsframe.graph)
            + "\n"
            + self.statsframe.dataframe
        )

        return s

    @staticmethod
    def thicketize_graphframe(gf, prf):
        """Necessary function to handle output from using GraphFrame readers.

        Arguments:
            gf (GraphFrame): hatchet GraphFrame object
            prf (str): profile source of the GraphFrame

        Returns:
            (thicket): Thicket object
        """
        th = Thicket(
            graph=gf.graph,
            dataframe=gf.dataframe,
            exc_metrics=gf.exc_metrics,
            inc_metrics=gf.inc_metrics,
            metadata=gf.metadata,
        )
        if th.profile is None and isinstance(prf, str):
            # Store used profiles and profile mappings using a hash of their string
            hash_arg = hash(prf)
            th.profile = [hash_arg]
            th.profile_mapping = OrderedDict({hash_arg: [prf]})
            # format metadata as a dict of dicts
            temp_meta = {}
            temp_meta[hash_arg] = th.metadata
            th.metadata = pd.DataFrame.from_dict(temp_meta, orient="index")
            th.metadata.index.set_names("profile", inplace=True)

            # Add profile to dataframe index
            th.dataframe["profile"] = hash_arg
            index_names = list(th.dataframe.index.names)
            index_names.insert(1, "profile")
            th.dataframe.reset_index(inplace=True)
            th.dataframe.set_index(index_names, inplace=True)
        return th

    @staticmethod
    def from_caliper(filename_or_stream, query=None, intersection=False):
        """Read in a Caliper .cali or .json file.

        Arguments:
            filename_or_stream (str or file-like): name of a Caliper output file in
                `.cali` or JSON-split format, or an open file object to read one
            query (str): cali-query in CalQL format
            intersection (bool): whether to perform intersection or union (default)
        """
        return Thicket.reader_dispatch(
            GraphFrame.from_caliper, intersection, filename_or_stream, query
        )

    @staticmethod
    def from_hpctoolkit(dirname, intersection=False):
        """Create a GraphFrame using hatchet's HPCToolkit reader and use its attributes
        to make a new thicket.

        Arguments:
            dirname (str): parent directory of an HPCToolkit experiment.xml file
            intersection (bool): whether to perform intersection or union (default)

        Returns:
            (thicket): new thicket containing HPCToolkit profile data
        """
        return Thicket.reader_dispatch(
            GraphFrame.from_hpctoolkit, intersection, dirname
        )

    @staticmethod
    def from_caliperreader(filename_or_caliperreader, intersection=False):
        """Helper function to read one caliper file.

        Arguments:
            filename_or_caliperreader (str or CaliperReader): name of a Caliper output
                file in `.cali` format, or a CaliperReader object
            intersection (bool): whether to perform intersection or union (default)
        """
        return Thicket.reader_dispatch(
            GraphFrame.from_caliperreader, intersection, filename_or_caliperreader
        )

    @staticmethod
    def reader_dispatch(func, intersection=False, *args, **kwargs):
        """Create a thicket from a list, directory of files, or a single file.

        Arguments:
            func (function): reader function to be used
            intersection (bool): whether to perform intersection or union (default).
            args (list): list of args; args[0] should be an object that can be read from
        """
        ens_list = []
        obj = args[0]  # First arg should be readable object

        # Parse the input object
        # if a list of files
        if type(obj) == list:
            for file in obj:
                ens_list.append(Thicket.thicketize_graphframe(func(file), file))
        # if directory of files
        elif os.path.isdir(obj):
            for file in os.listdir(obj):
                f = os.path.join(obj, file)
                ens_list.append(Thicket.thicketize_graphframe(func(f), f))
        # if single file
        elif os.path.isfile(obj):
            return Thicket.thicketize_graphframe(func(*args, **kwargs), args[0])
        else:
            raise TypeError(type(obj) + " is not a valid type to be read from.")

        # Perform unify ensemble
        thicket_object = Thicket.unify_ensemble(ens_list)
        if intersection:
            thicket_object.intersection()
        return thicket_object

    @staticmethod
    def from_json(json_thicket):
        # deserialize the json
        thicket_dict = json.loads(json_thicket)
        hatchet_spec = {
            "graph": thicket_dict["graph"],
            "dataframe": thicket_dict["dataframe"],
            "dataframe_indices": thicket_dict["dataframe_indices"],
            "exclusive_metrics": thicket_dict["inclusive_metrics"],
            "inclusive_metrics": thicket_dict["exclusive_metrics"],
        }

        gf = GraphFrame.from_json(json.dumps(hatchet_spec))

        th = Thicket(
            graph=gf.graph,
            dataframe=gf.dataframe,
            exc_metrics=thicket_dict["exclusive_metrics"],
            inc_metrics=thicket_dict["inclusive_metrics"],
            profile=thicket_dict["profile"],
            profile_mapping=thicket_dict["profile_mapping"],
        )

        if "metadata" in thicket_dict:
            mf = pd.DataFrame(thicket_dict["metadata"])
            mf.set_index(mf["profile"], inplace=True)
            if "profile" in mf.columns:
                mf = mf.drop(columns=["profile"])
            th.metadata = mf

        # catch condition where there are no stats
        if "stats" in thicket_dict:
            sf_spec = {
                "graph": thicket_dict["graph"],
                "dataframe": thicket_dict["stats"],
                "dataframe_indices": ["node"],
                "exclusive_metrics": [],
                "inclusive_metrics": [],
            }
            th.statsframe = GraphFrame.from_json(json.dumps(sf_spec))
            th.statsframe.graph = th.graph

        # make and return thicket?
        return th

    @staticmethod
    def columnar_join(
        thicket_list,
        header_list=None,
        column_name=None,
    ):
        """Join Thickets column-wise. New column multi-index will be created with
        columns under separate indexer headers.

        Arguments:
            thicket_list (list): List of Thickets to join
            header_list (list): List of headers to use for the new columnar multi-index
            column_name (str): Name of the column from the metadata table to join on. If
                no argument is provided, it is assumed that there is no profile-wise
                relationship between self and other.

        Returns:
            (Thicket): New Thicket object with joined columns
        """

        def _create_multiindex_columns(df, upper_idx_name):
            """Helper function to create multi-index column names from a dataframe's
            current columns.

            Arguments:
            df (DataFrame): source dataframe
            upper_idx_name (String): name of the newly added index in the multi-index.
                Prepended before each column as a tuple.

            Returns:
                (list): list of new indicies generated from the source dataframe
            """
            new_idx = []
            for column in df.columns:
                new_tuple = (upper_idx_name, column)
                new_idx.append(new_tuple)
            return new_idx

        ###
        # Step 0A: Pre-check of data structures
        ###
        # Required/expected format of the data
        for th in thicket_list:
            verify_thicket_structures(th.dataframe, index=["node", "profile"])
            verify_thicket_structures(th.statsframe.dataframe, index=["node"])
            verify_thicket_structures(th.metadata, index=["profile"])
        # Check for column_name in metadata
        if column_name:
            for th in thicket_list:
                verify_thicket_structures(th.metadata, columns=[column_name])
        # Check length of profiles match
        for i in range(len(thicket_list) - 1):
            if len(thicket_list[i].profile) != len(thicket_list[i + 1].profile):
                raise ValueError(
                    "Length of all thicket profiles must match. {} != {}".format(
                        len(thicket_list[i].profile), len(thicket_list[i + 1].profile)
                    )
                )
        # Ensure all thickets profiles are sorted. Must be true when column_name=None to
        # guarantee performance data table and metadata table match up.
        if column_name is None:
            for th in thicket_list:
                verify_sorted_profile(th.dataframe)
                verify_sorted_profile(th.metadata)

        ###
        # Step 0B: Variable Initialization
        ###
        # Initialize combined thicket
        combined_th = thicket_list[0].deepcopy()
        # Use copies to be non-destructive
        thicket_list_cp = [th.deepcopy() for th in thicket_list]

        ###
        # Step 1: Unify the thickets
        ###
        # Unify graphs if "self" and "other" do not have the same graph
        union_graph = thicket_list_cp[0].graph
        for i in range(len(thicket_list_cp) - 1):
            if thicket_list_cp[i].graph != thicket_list_cp[i + 1].graph:
                union_graph = union_graph.union(thicket_list_cp[i + 1].graph)
        combined_th.graph = union_graph
        for i in range(len(thicket_list_cp)):
            # Set all graphs to the union graph
            thicket_list_cp[i].graph = union_graph
            # Necessary to change dataframe hatchet id's to match the nodes in the graph
            helpers._sync_nodes_frame(union_graph, thicket_list_cp[i].dataframe)
            # For tree diff. dataframes need to be sorted.
            thicket_list_cp[i].dataframe.sort_index(inplace=True)

        ###
        # Step 2: Join "self" & "other" performance data table
        ###
        # Create header list if not provided
        if header_list is None:
            header_list = [i for i in range(len(thicket_list))]

        # Update index to reflect performance data table index
        new_mappings = {}  # Dictionary mapping old profiles to new profiles
        if column_name is None:  # Create index from scratch
            new_profiles = [i for i in range(len(thicket_list_cp[0].profile))]
            for i in range(len(thicket_list_cp)):
                thicket_list_cp[i].metadata["new_profiles"] = new_profiles
                thicket_list_cp[i].add_column_from_metadata_to_ensemble(
                    "new_profiles", drop=True
                )
                thicket_list_cp[i].dataframe.reset_index(level="profile", inplace=True)
                new_mappings.update(
                    pd.Series(
                        thicket_list_cp[i]
                        .dataframe["new_profiles"]
                        .map(lambda x: (x, header_list[i]))
                        .values,
                        index=thicket_list_cp[i].dataframe["profile"],
                    ).to_dict()
                )
                thicket_list_cp[i].dataframe.drop("profile", axis=1, inplace=True)
                thicket_list_cp[i].dataframe.set_index(
                    "new_profiles", append=True, inplace=True
                )
                thicket_list_cp[i].dataframe.index.rename(
                    "profile", level="new_profiles", inplace=True
                )
        else:  # Change second-level index to be from metadata's "column_name" column
            for i in range(len(thicket_list_cp)):
                thicket_list_cp[i].add_column_from_metadata_to_ensemble(column_name)
                thicket_list_cp[i].dataframe.reset_index(level="profile", inplace=True)
                new_mappings.update(
                    pd.Series(
                        thicket_list_cp[i]
                        .dataframe[column_name]
                        .map(lambda x: (x, header_list[i]))
                        .values,
                        index=thicket_list_cp[i].dataframe["profile"],
                    ).to_dict()
                )
                thicket_list_cp[i].dataframe.drop("profile", axis=1, inplace=True)
                thicket_list_cp[i].dataframe.set_index(
                    column_name, append=True, inplace=True
                )
                thicket_list_cp[i].dataframe.sort_index(inplace=True)

        # Create tuple columns
        new_columns = [
            _create_multiindex_columns(th.dataframe, header_list[i])
            for i, th in enumerate(thicket_list_cp)
        ]
        # Clear old metrics (non-tuple)
        combined_th.exc_metrics.clear()
        combined_th.inc_metrics.clear()
        # Update inc/exc metrics
        for i in range(len(new_columns)):
            for col_tuple in new_columns[i]:
                if col_tuple[1] in thicket_list_cp[i].exc_metrics:
                    combined_th.exc_metrics.append(col_tuple)
                if col_tuple[1] in thicket_list_cp[i].inc_metrics:
                    combined_th.inc_metrics.append(col_tuple)
        # Update columns
        for i in range(len(thicket_list_cp)):
            thicket_list_cp[i].dataframe.columns = pd.MultiIndex.from_tuples(
                new_columns[i]
            )

        # Concat performance data table together
        combined_th.dataframe = pd.concat(
            [thicket_list_cp[i].dataframe for i in range(len(thicket_list_cp))],
            axis="columns",
            join="outer",
        )

        # Extract "name" columns to upper level
        nodes = list(set(combined_th.dataframe.reset_index()["node"]))
        for node in nodes:
            combined_th.dataframe.loc[node, "name"] = node.frame["name"]
        combined_th.dataframe.drop(
            columns=[(header_list[i], "name") for i in range(len(header_list))],
            inplace=True,
        )

        ###
        # Step 3: Join "self" & "other" metadata table
        ###
        # Update index to reflect performance data table index
        for i in range(len(thicket_list_cp)):
            thicket_list_cp[i].metadata.reset_index(drop=True, inplace=True)
        if column_name is None:
            for i in range(len(thicket_list_cp)):
                thicket_list_cp[i].metadata.index.set_names("profile", inplace=True)
        else:
            for i in range(len(thicket_list_cp)):
                thicket_list_cp[i].metadata.set_index(column_name, inplace=True)
                thicket_list_cp[i].metadata.sort_index(inplace=True)

        # Create multi-index columns
        for i in range(len(thicket_list_cp)):
            thicket_list_cp[i].metadata.columns = pd.MultiIndex.from_tuples(
                _create_multiindex_columns(thicket_list_cp[i].metadata, header_list[i])
            )

        # Concat metadata together
        combined_th.metadata = pd.concat(
            [thicket_list_cp[i].metadata for i in range(len(thicket_list_cp))],
            axis="columns",
            join="outer",
        )

        ###
        # Step 4: Update other Thicket objects.
        ###
        for i in range(1, len(thicket_list_cp)):
            combined_th.profile += thicket_list_cp[i].profile  # Update "profile" object
            combined_th.profile_mapping.update(
                thicket_list_cp[i].profile_mapping
            )  # Update "profile_mapping" object
        combined_th.profile = [new_mappings[prf] for prf in combined_th.profile]
        profile_mapping_cp = combined_th.profile_mapping.copy()
        for k, v in profile_mapping_cp.items():
            combined_th.profile_mapping[
                new_mappings[k]
            ] = combined_th.profile_mapping.pop(k)

        # Clear aggregated statistics table
        combined_th.statsframe = GraphFrame(
            graph=combined_th.graph,
            dataframe=helpers._new_statsframe_df(
                combined_th.dataframe, multiindex=True
            ),
        )
        combined_th.performance_cols = helpers._get_perf_columns(combined_th.dataframe)

        return combined_th

    def add_column_from_metadata_to_ensemble(
        self, column_name, overwrite=False, drop=False
    ):
        """Add a column from the metadata table to the performance data table.

        Arguments:
            column_name (str): Name of the column from the metadata table
            overwrite (bool): Determines overriding behavior in performance data table
            drop (bool): Whether to drop the column from the metadata table afterwards
        """
        # Add warning if column already exists in performance data table
        if column_name in self.dataframe.columns:
            # Drop column to overwrite, otherwise warn and return
            if overwrite:
                self.dataframe.drop(column_name, axis=1, inplace=True)
            else:
                warnings.warn(
                    "Column "
                    + column_name
                    + " already exists. Set 'overwrite=True' to force update the column."
                )
                return

        # Add the column to the performance data table
        self.dataframe = self.dataframe.join(
            self.metadata[column_name], on=self.dataframe.index.names[1]
        )

        # Drop column
        if drop:
            self.metadata.drop(column_name, axis=1, inplace=True)

    def squash(self, update_inc_cols=True):
        """Rewrite the Graph to include only nodes present in the performance
        data table's rows.

        This can be used to simplify the Graph, or to normalize Graph indexes between
        two Thickets.

        Arguments:
            update_inc_cols (boolean, optional): if True, update inclusive columns.

        Returns:
            (Thicket): a newly squashed Thicket object
        """
        squashed_gf = GraphFrame.squash(self, update_inc_cols=update_inc_cols)
        new_graph = squashed_gf.graph
        # The following code updates the performance data and the aggregated statistics
        # table with the remaining (re-indexed) nodes. The dataframe is internally
        # updated in squash(), so we can easily just save it to our thicket performance
        # data. For the aggregated statistics table, we'll have to come up with a better
        # way eventually, but for now, we'll just create a new aggregated statistics
        # table the same way we do when we create a new thicket.
        new_dataframe = squashed_gf.dataframe
        stats_df = helpers._new_statsframe_df(new_dataframe)
        sframe = GraphFrame(
            graph=new_graph,
            dataframe=stats_df,
        )
        return Thicket(
            new_graph,
            new_dataframe,
            exc_metrics=self.exc_metrics,
            inc_metrics=self.inc_metrics,
            default_metric=self.default_metric,
            metadata=self.metadata,
            profile=self.profile,
            profile_mapping=self.profile_mapping,
            statsframe=sframe,
        )

    def copy(self):
        """Return a partially shallow copy of the Thicket.

        See GraphFrame.copy() for more details

        Arguments:
            self (Thicket): object to make a copy of

        Returns:
            other (Thicket): copy of self
                (graph ... default_metric): Same behavior as GraphFrame
                metadata (DataFrame): pandas "non-deep" copy of dataframe
                profile (list): copy of self's profile
                profile_mapping (dict): copy of self's profile_mapping
                statsframe (GraphFrame): calls GraphFrame.copy()
        """
        gf = GraphFrame.copy(self)

        return Thicket(
            graph=gf.graph,
            dataframe=gf.dataframe,
            exc_metrics=gf.exc_metrics,
            inc_metrics=gf.inc_metrics,
            default_metric=gf.default_metric,
            metadata=self.metadata.copy(deep=False),
            profile=copy.copy(self.profile),
            profile_mapping=copy.copy(self.profile_mapping),
            statsframe=self.statsframe.copy(),
        )

    def deepcopy(self):
        """Return a deep copy of the Thicket.

        See GraphFrame.deepcopy() for more details

        Arguments:
            self (Thicket): object to make a copy of

        Returns:
            other (Thicket): copy of self
                (graph ... default_metric): same behavior as GraphFrame
                metadata (DataFrame): pandas "deep" copy of dataframe
                profile (list): copy of self's profile
                profile_mapping (dict): copy of self's profile_mapping
                statsframe (GraphFrame): Calls GraphFrame.deepcopy()
        """
        gf = GraphFrame.deepcopy(self)

        return Thicket(
            graph=gf.graph,
            dataframe=gf.dataframe,
            exc_metrics=gf.exc_metrics,
            inc_metrics=gf.inc_metrics,
            default_metric=gf.default_metric,
            metadata=self.metadata.copy(),
            profile=copy.deepcopy(self.profile),
            profile_mapping=copy.deepcopy(self.profile_mapping),
            statsframe=self.statsframe.deepcopy(),
        )

    def tree(self):
        """hatchet tree() function for a thicket"""
        temp_df = self.statsframe.dataframe.copy()
        # Adjustments specific for multi-index.
        if isinstance(temp_df.columns, pd.MultiIndex):
            temp_df.columns = temp_df.columns.to_flat_index()
            temp_df.rename(columns={("", "name"): "name"}, inplace=True)
        # Placeholder value. TODO: Enable selection from performance data table.
        temp_df["thicket_tree"] = -1
        return GraphFrame.tree(
            self=Thicket(graph=self.graph, dataframe=temp_df),
            metric_column="thicket_tree",
        )

    def unify_pair(self, other):
        """Unify two Thicket's graphs and dataframes"""
        # Check for the same object. Cheap operation since no graph walkthrough.
        if self.graph is other.graph:
            print("same graph (object)")
            return self.graph

        # Check for the same graph structure. Need to walk through graphs *but should
        # still be less expensive then performing the rest of this function.*
        if self.graph == other.graph:
            print("same graph (structure)")
            return self.graph

        print("different graph")

        node_map = {}
        union_graph = self.graph.union(other.graph, node_map)

        self_index_names = self.dataframe.index.names
        other_index_names = other.dataframe.index.names

        self.dataframe.reset_index(inplace=True)
        other.dataframe.reset_index(inplace=True)

        self.dataframe["node"] = self.dataframe["node"].apply(lambda x: node_map[id(x)])
        other.dataframe["node"] = other.dataframe["node"].apply(
            lambda x: node_map[id(x)]
        )

        self.dataframe.set_index(self_index_names, inplace=True)
        other.dataframe.set_index(other_index_names, inplace=True)

        self.graph = union_graph
        other.graph = union_graph

        return union_graph

    @staticmethod
    def unify_pairwise(th_list, debug=False):
        """Unifies two thickets graphs and dataframes.

        Ensure self and other have the same graph and same node IDs. This may change the
        node IDs in the dataframe.

        Update the graphs in the graphframe if they differ.

        Arguments:
            th_list (list): list of Thicket objects
            debug (bool): print debug statements

        Returns:
            union_graph (Graph): unified graph
        """
        union_graph = th_list[0].graph
        for i in range(len(th_list)):
            for j in range(i + 1, len(th_list)):
                if debug:
                    print("Unifying (" + str(i) + ", " + str(j) + "...")
                union_graph = th_list[i].unify_pair(th_list[j])
        return union_graph

    @staticmethod
    def unify_listwise(th_list, debug=False):
        """Unify a list of Thicket's graphs and dataframes

        Arguments:
            th_list (list): list of Thicket objects
            debug (bool): print debug statements

        Returns:
            union_graph (Graph): unified graph
        """
        # variable to keep track of case where all graphs are the same
        same_graphs = True

        # GRAPH UNIFICATION
        union_graph = th_list[0].graph
        for i in range(1, len(th_list)):  # n-1 unions
            # Check to skip unnecessary computation. apply short circuiting with 'or'.
            if union_graph is th_list[i].graph or union_graph == th_list[i].graph:
                if debug:
                    print("Union Graph == thicket[" + str(i) + "].graph")
            else:
                if debug:
                    print("Unifying (Union Graph, " + str(i) + ")")
                same_graphs = False
                # Unify graph with current thickets graph
                union_graph = union_graph.union(th_list[i].graph)

        # If the graphs were all the same in the first place then there is no need to
        # apply any node mappings.
        if same_graphs:
            return union_graph

        # DATAFRAME MAPPING UPDATE
        for i in range(len(th_list)):  # n ops
            node_map = {}
            # Create a node map from current thickets graph to the union graph. This is
            # only valid once the union graph is complete.
            union_graph.union(th_list[i].graph, node_map)
            names = th_list[i].dataframe.index.names
            th_list[i].dataframe.reset_index(inplace=True)

            # Apply node_map mapping
            th_list[i].dataframe["node"] = (
                th_list[i].dataframe["node"].apply(lambda node: node_map[id(node)])
            )
            th_list[i].dataframe.set_index(names, inplace=True, drop=True)

        # After this point the graph and dataframe in each thicket is out of sync.
        # We could update the graph element in thicket to be the union graph but if the
        # user prints out the graph how do we annotate nodes only contained in one
        # thicket.
        return union_graph

    @staticmethod
    def unify_ensemble(th_list, pairwise=False, superthicket=False):
        """Unify a list of thickets into a single thicket

        Arguments:
            th_list (list): list of thickets
            pairwise (bool): use the pairwise implementation of unify (use if having
                issues)
            superthicket (bool): whether the result is a superthicket

        Returns:
            (thicket): unified thicket
        """
        unify_graph = None
        if pairwise:
            unify_graph = Thicket.unify_pairwise(th_list)
        else:
            unify_graph = Thicket.unify_listwise(th_list)

        helpers._resolve_missing_indicies(th_list)

        # Unify dataframe
        unify_df = pd.DataFrame()
        unify_inc_metrics = []
        unify_exc_metrics = []
        unify_metadata = pd.DataFrame()
        unify_profile = []
        unify_profile_mapping = {}

        # Unification loop
        for i, th in enumerate(th_list):
            unify_inc_metrics.extend(th.inc_metrics)
            unify_exc_metrics.extend(th.exc_metrics)
            if len(th.metadata) > 0:
                curr_meta = th.metadata.copy()
                unify_metadata = pd.concat([curr_meta, unify_metadata])
            if th.profile is not None:
                unify_profile.extend(th.profile)
            if th.profile_mapping is not None:
                unify_profile_mapping.update(th.profile_mapping)
            unify_df = pd.concat([th.dataframe, unify_df])

        # Operations specific to a superthicket
        if superthicket:
            unify_metadata.index.rename("thicket", inplace=True)

            # Process to aggregate rows of thickets with the same name.
            def _agg_function(obj):
                """Aggregate values in 'obj' into a set to remove duplicates."""
                if len(obj) <= 1:
                    return obj
                else:
                    _set = set(obj)
                    if len(_set) == 1:
                        return _set.pop()
                    else:
                        return _set

            unify_metadata = unify_metadata.groupby("thicket").agg(_agg_function)

        # Have metadata index match performance data table index
        unify_metadata.sort_index(inplace=True)

        # Sort by hatchet node id
        unify_df.sort_index(inplace=True)

        unify_inc_metrics = list(set(unify_inc_metrics))
        unify_exc_metrics = list(set(unify_exc_metrics))

        # Workaround for graph/df node id mismatch.
        # (n tree nodes) X (m df nodes) X (m)
        helpers._sync_nodes(unify_graph, unify_df)

        # Mutate into OrderedDict to sort profile hashes
        unify_profile_mapping = OrderedDict(sorted(unify_profile_mapping.items()))

        unify_th = Thicket(
            graph=unify_graph,
            dataframe=unify_df,
            exc_metrics=unify_exc_metrics,
            inc_metrics=unify_inc_metrics,
            metadata=unify_metadata,
            profile=unify_profile,
            profile_mapping=unify_profile_mapping,
        )
        return unify_th

    @staticmethod
    def make_superthicket(th_list, profiles_from_meta=None):
        """Convert a list of thickets into a 'superthicket'.

        Their individual aggregated statistics table are ensembled and become the
        superthicket's performance data table.

        Arguments:
            th_list (list): list of thickets
            profiles_from_meta (str, optional): name of the metadata column to use as
                the new second-level index. Uses the first value so this only makes
                sense if provided column is all equal values and each thicket's columns
                differ in value.

        Returns:
            (thicket): superthicket
        """
        # Pre-check of data structures
        for th in th_list:
            verify_thicket_structures(
                th.dataframe, index=["node", "profile"]
            )  # Required for deepcopy operation
            verify_thicket_structures(
                th.statsframe.dataframe, index=["node"]
            )  # Required for deepcopy operation

        # Setup names list
        th_names = []
        if profiles_from_meta is None:
            for i in range(len(th_list)):
                th_names.append(i)
        else:  # profiles_from_meta was provided.
            for th in th_list:
                # Get name from metadata table
                name_list = th.metadata[profiles_from_meta].tolist()

                if len(name_list) > 1:
                    warnings.warn(
                        f"Multiple values for name {name_list} at thicket.metadata[{profiles_from_meta}]. Only the first will be used."
                    )
                th_names.append(name_list[0])

        th_copy_list = []
        for i in range(len(th_list)):
            th_copy = th_list[i].deepcopy()

            th_id = th_names[i]

            # Modify graph
            # Necessary so node ids match up
            th_copy.graph = th_copy.statsframe.graph

            # Modify the performance data table
            df = th_copy.statsframe.dataframe
            df["thicket"] = th_id
            df.set_index("thicket", inplace=True, append=True)
            th_copy.dataframe = df

            # Adjust profile and profile_mapping
            th_copy.profile = [th_id]
            profile_paths = list(th_copy.profile_mapping.values())
            th_copy.profile_mapping = OrderedDict({th_id: profile_paths})

            # Modify metadata dataframe
            idx_name = "new_idx"
            th_copy.metadata[idx_name] = th_id
            th_copy.metadata.set_index(idx_name, inplace=True)

            # Append copy to list
            th_copy_list.append(th_copy)

        return Thicket.unify_ensemble(th_copy_list, superthicket=True)

    def to_json(self, ensemble=True, metadata=True, stats=True):
        jsonified_thicket = {}

        # jsonify graph
        """
        Nodes: {hatchet_nid: {node data, children:[by-id]}}
        """
        formatted_graph_dict = {}

        for n in self.graph.traverse():
            formatted_graph_dict[n._hatchet_nid] = {
                "data": n.frame.attrs,
                "children": [c._hatchet_nid for c in n.children],
                "parents": [c._hatchet_nid for c in n.parents],
            }

        jsonified_thicket["graph"] = [formatted_graph_dict]

        if ensemble:
            jsonified_thicket["dataframe_indices"] = list(self.dataframe.index.names)
            ef = self.dataframe.reset_index()
            ef["node"] = ef["node"].apply(lambda n: n._hatchet_nid)
            jsonified_thicket["dataframe"] = ef.replace({np.nan: None}).to_dict(
                "records"
            )
        if metadata:
            jsonified_thicket["metadata"] = (
                self.metadata.reset_index().replace({np.nan: None}).to_dict("records")
            )
        if stats:
            sf = self.statsframe.dataframe.copy(deep=True)
            sf.index = sf.index.map(lambda n: n._hatchet_nid)
            jsonified_thicket["stats"] = (
                sf.replace({np.nan: None}).reset_index().to_dict("records")
            )

        jsonified_thicket["inclusive_metrics"] = self.inc_metrics
        jsonified_thicket["exclusive_metrics"] = self.exc_metrics
        jsonified_thicket["profile"] = self.profile
        jsonified_thicket["profile_mapping"] = self.profile_mapping

        return json.dumps(jsonified_thicket)

    def intersection(self):
        """Perform an intersection operation on a thicket.

        Nodes not contained in all profiles are removed.

        Returns:
            remaining_node_list (list): list of nodes that were not removed
            removed_node_list (list): list of removed nodes
        """
        # Filter the performance data table
        total_profiles = len(self.profile)
        remaining_node_list = []  # Needed for graph
        removed_node_list = []

        # For each node
        for node, new_df in self.dataframe.groupby(level=0):
            # Use profile count to make decision
            if len(new_df) < total_profiles:
                removed_node_list.append(node)
            else:
                remaining_node_list.append(node)
        self.dataframe.drop(removed_node_list, inplace=True)

        # Propagate change to aggregated statistics table
        self.statsframe.dataframe.drop(removed_node_list, inplace=True)

        # Filter the graph

        # Remove roots
        self.graph.roots = list(set(self.graph.roots).intersection(remaining_node_list))
        for node in self.graph.traverse():
            # Remove children and parents that DNE in the intersection
            new_children = []
            new_parents = []
            for child in node.children:
                if child in remaining_node_list:
                    new_children.append(child)
            for parent in node.parents:
                if parent in remaining_node_list:
                    new_parents.append(parent)
            node.children = new_children
            node.parents = new_parents

        # Update hatchet nids
        self.graph.enumerate_traverse()

        return remaining_node_list, removed_node_list

    def filter_metadata(self, select_function):
        """Filter thicket object based on a metadata key.

        Changes are propagated to the entire thicket object.

        Arguments:
            select_function (lambda function): filter to apply to the metadata table

        Returns:
            (thicket): new thicket object with selected value
        """
        if callable(select_function):
            if not self.metadata.empty:
                # check only 1 index in metadata
                assert self.metadata.index.nlevels == 1

                # Add warning if filtering on multi-index columns
                if isinstance(self.metadata.columns, pd.MultiIndex):
                    warnings.warn(
                        "Filtering on MultiIndex columns will impact the entire row, not just the subsection of the provided MultiIndex."
                    )

                # Get index name
                index_name = self.metadata.index.name

                # create a copy of the thicket object
                new_thicket = self.copy()

                # filter metadata table
                filtered_rows = new_thicket.metadata.apply(select_function, axis=1)
                new_thicket.metadata = new_thicket.metadata[filtered_rows]

                # note index keys to filter performance data table
                index_id = new_thicket.metadata.index.values.tolist()
                # filter performance data table based on the metadata table
                new_thicket.dataframe = new_thicket.dataframe[
                    new_thicket.dataframe.index.get_level_values(index_name).isin(
                        index_id
                    )
                ]

                # create an empty aggregated statistics table with the name column
                new_thicket.statsframe.dataframe = helpers._new_statsframe_df(
                    new_thicket.dataframe
                )
            else:
                raise EmptyMetadataTable(
                    "The provided Thicket object has an empty MetadataTable."
                )

        else:
            raise InvalidFilter("The argument passed to filter must be a callable.")

        return new_thicket

    def filter(self, filter_func):
        """Overloaded generic filter function.

        Provides a helper message for using the thicket filter functions.
        """
        raise RuntimeError(
            "Invalid function: thicket.filter(), please use thicket.filter_metadata() or thicket.filter_stats()"
        )

    def query(self, query_obj, squash=True, update_inc_cols=True):
        """Apply a Hatchet query to the Thicket object.

        Arguments:
            query_obj (AbstractQuery): the query, represented as by a subclass of
                Hatchet's AbstractQuery
            squash (bool): if true, run Thicket.squash before returning the result of
                the query
            update_inc_cols (boolean, optional): if True, update inclusive columns when
                performing squash.

        Returns:
            (Thicket): a new Thicket object containing the data that matches the query
        """
        if isinstance(query_obj, (list, str)):
            raise UnsupportedQuery(
                "Object and String queries from Hatchet are not yet supported in Thicket"
            )
        elif not issubclass(type(query_obj), AbstractQuery):
            raise TypeError(
                "Input to 'query' must be a Hatchet query (i.e., list, str, or subclass of AbstractQuery)"
            )
        dframe_copy = self.dataframe.copy()
        index_names = self.dataframe.index.names
        dframe_copy.reset_index(inplace=True)
        query = query_obj
        # TODO Add a conditional here to parse Object and String queries when supported
        query_matches = query.apply(self)
        filtered_df = dframe_copy.loc[dframe_copy["node"].isin(query_matches)]
        if filtered_df.shape[0] == 0:
            raise EmptyQuery("The provided query would have produced an empty Thicket.")
        filtered_df.set_index(index_names, inplace=True)
        filtered_th = Thicket(
            self.graph,
            filtered_df,
            exc_metrics=self.exc_metrics,
            inc_metrics=self.inc_metrics,
            default_metric=self.default_metric,
            metadata=self.metadata,
            profile=self.profile,
            profile_mapping=self.profile_mapping,
            statsframe=self.statsframe,
        )
        if squash:
            return filtered_th.squash(update_inc_cols=update_inc_cols)
        return filtered_th

    def groupby(self, groupby_function):
        """Create sub-thickets based on unique values in metadata column(s).

        Arguments:
            groupby_function (groupby_function): groupby function on dataframe

        Returns:
            (list): list of (sub)thickets
        """
        if not self.metadata.empty:
            # group metadata table by unique values in a column
            sub_metadataframes = self.metadata.groupby(groupby_function, dropna=False)

            list_sub_thickets = []
            unique_vals = []

            # for all unique groups of metadata table
            for key, df in sub_metadataframes:
                unique_vals.append(key)

                # create a thicket copy
                sub_thicket = self.copy()

                # return unique group as the metadata table
                sub_thicket.metadata = df

                # find profiles in current unique group and filter performance data
                # table
                profile_id = df.index.values.tolist()
                sub_thicket.dataframe = sub_thicket.dataframe[
                    sub_thicket.dataframe.index.get_level_values("profile").isin(
                        profile_id
                    )
                ]

                # Updates the profiles to only contain the remaining ones
                profile_mapping_tmp = sub_thicket.profile_mapping.copy()
                for profile_mapping_key in profile_mapping_tmp:
                    if profile_mapping_key not in profile_id:
                        sub_thicket.profile_mapping.pop(profile_mapping_key)

                sub_thicket.profile = [
                    profile for profile in sub_thicket.profile if profile in profile_id
                ]

                # clear the aggregated statistics table for current unique group
                sub_thicket.statsframe.dataframe = helpers._new_statsframe_df(
                    sub_thicket.dataframe
                )
                list_sub_thickets.append(sub_thicket)
        else:
            raise EmptyMetadataTable(
                "The provided Thicket object has an empty metadata table."
            )

        print(len(list_sub_thickets), " thickets created...")
        print(unique_vals)

        return list_sub_thickets

    def filter_stats(self, filter_function):
        """Filter thicket object based on a stats column.

        Propagate changes to the entire thicket object.

        Arguments:
            select_function (lambda function): filter to apply to the aggregated
                statistics table

        Returns:
            (thicket): new thicket object with applied filter function
        """
        # copy thicket
        new_thicket = self.copy()

        # filter aggregated statistics table based on greater than restriction
        filtered_rows = new_thicket.statsframe.dataframe.apply(filter_function, axis=1)
        new_thicket.statsframe.dataframe = new_thicket.statsframe.dataframe[
            filtered_rows
        ]

        # filter performance data table based on filtered nodes
        filtered_nodes = new_thicket.statsframe.dataframe.index.values.tolist()
        new_thicket.dataframe = new_thicket.dataframe[
            new_thicket.dataframe.index.get_level_values("node").isin(filtered_nodes)
        ]

        # filter nodes in the graphframe based on the dataframe nodes
        # TODO see if the new Thicket.squash function will work here
        filtered_graphframe = GraphFrame.squash(new_thicket)
        new_thicket.graph = filtered_graphframe.graph
        new_thicket.statsframe.graph = filtered_graphframe.graph

        return new_thicket

    def get_unique_metadata(self):
        """Get unique values per column in metadata.

        Returns:
            (dict): alphabetical ordered dictionary with key's being the column names
                and the values being unique values for a metadata column.
        """
        unique_meta = {}

        # thicket object without columnar index
        if self.dataframe.columns.nlevels == 1:
            for col in self.metadata.columns:
                # skip columns where the values are a list
                if isinstance(self.metadata[col].iloc[0], list):
                    continue
                else:
                    unique_entries = self.metadata[col].unique().tolist()
                    unique_meta[col] = unique_entries

            sorted_meta = dict(sorted(unique_meta.items(), key=lambda x: x[0].lower()))
        # columnar joined thicket object
        else:
            sorted_meta = []
            for idx in list(self.metadata.columns.levels[0]):
                for col in self.metadata[idx].columns:
                    if isinstance(self.metadata[idx][col].iloc[0], list):
                        continue
                    else:
                        unique_entries = self.metadata[idx][col].unique().tolist()
                        unique_meta[col] = unique_entries

                sorted_meta.append(
                    (idx, dict(sorted(unique_meta.items(), key=lambda x: x[0].lower())))
                )

        return sorted_meta


class InvalidFilter(Exception):
    """Raised when an invalid argument is passed to the filter function."""


class EmptyMetadataTable(Exception):
    """Raised when a Thicket object argument is passed with an empty MetadataTable to
    the filter function.
    """


class UnsupportedQuery(Exception):
    """Raised when an object query or string query are provided to the 'query' function
    because those types of queries are not yet supported in Thicket.
    """


class EmptyQuery(Exception):
    """Raised when a query would result in an empty Thicket object."""



# Copyright 2022 Lawrence Livermore National Security, LLC and other
# Thicket Project Developers. See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT

# make flake8 unused names in this file.
# flake8: noqa: F401

from .thicket import Thicket
from .thicket import InvalidFilter
from .thicket import EmptyMetadataTable
from .stats.maximum import maximum
from .stats.mean import mean
from .stats.median import median
from .stats.minimum import minimum
from .stats.percentiles import percentiles
from .stats.std import std
from .stats.variance import variance
from .stats.calc_boxplot_statistics import calc_boxplot_statistics
from .stats.correlation_nodewise import correlation_nodewise
from .stats.check_normality import check_normality
from .stats.display_boxplot import display_boxplot
from .stats.display_histogram import display_histogram
from .stats.display_heatmap import display_heatmap
from .version import __version__



# Copyright 2022 Lawrence Livermore National Security, LLC and other
# Thicket Project Developers. See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT

import numpy as np
import pandas as pd

from ..utils import verify_thicket_structures


def std(thicket, columns=None):
    """Calculate the standard deviation for each node in the performance data table.

    Designed to take in a thicket, and append one or more columns to the aggregated
    statistics table for the standard deviation calculation for each node.

    Standard deviation describes how dispersed the data is in relation to the mean.

    Arguments:
        thicket (thicket): Thicket object
        columns (list): List of hardware/timing metrics to perform standard deviation
            calculation on. Note, if using a columnar_joined thicket a list of tuples
            must be passed in with the format (column index, column name).
    """
    if columns is None:
        raise ValueError(
            "To see a list of valid columns, please run Thicket.get_perf_columns()."
        )

    verify_thicket_structures(
        thicket.dataframe, index=["node", "profile"], columns=columns
    )

    # thicket object without columnar index
    if thicket.dataframe.columns.nlevels == 1:
        for column in columns:
            std = []
            for node in pd.unique(thicket.dataframe.reset_index()["node"].tolist()):
                std.append(np.std(thicket.dataframe.loc[node][column]))
            thicket.statsframe.dataframe[column + "_std"] = std
    # columnar joined thicket object
    else:
        for idx, column in columns:
            std = []
            for node in pd.unique(thicket.dataframe.reset_index()["node"].tolist()):
                std.append(np.std(thicket.dataframe.loc[node][(idx, column)]))
            thicket.statsframe.dataframe[(idx, column + "_std")] = std

        # sort columns in index
        thicket.statsframe.dataframe = thicket.statsframe.dataframe.sort_index(axis=1)



# Copyright 2022 Lawrence Livermore National Security, LLC and other
# Thicket Project Developers. See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT

import re
from collections import defaultdict

import pandas as pd
import ncu_report


def add_ncu_metrics(th, ncu_report_mapping, chosen_metrics=None):
    """Add metrics from one ncu report file to one thicket

    Arguments:
        th (Thicket): Thicket object
        ncu_report_mapping (dict): dictionary mapping from "NCU report file":"Thicket
            profile"
        chosen_metrics (str): If known, which NCU metrics to add
    """
    # Keep list of NCU dfs for concat
    ncu_df_list = []

    # Loop through dict
    for ncu_report_file in ncu_report_mapping:
        # Hash value that should exist in th
        ncu_hash = hash(ncu_report_mapping[ncu_report_file])

        # Load kernels
        report = ncu_report.load_report(ncu_report_file)
        kernels = report[0]

        # Mapping from node names to kernels
        nodes = th.dataframe.index.get_level_values("node").tolist()
        names = th.dataframe["name"]
        cpu_side_kernels = {}
        for i in range(len(nodes)):
            if nodes[i].frame["type"] == "kernel":
                cpu_side_kernels[names[i]] = nodes[i]

        # Pre-processing
        # Remove warmup kernels
        warmup_end_idx = 0
        i = 0
        first_warmup_kernel = kernels[0]
        same_kernel = True
        for i in range(1, len(kernels)):
            if kernels[i].name(
                kernels[i].NameBase_DEMANGLED
            ) != first_warmup_kernel.name(first_warmup_kernel.NameBase_DEMANGLED):
                same_kernel = False
            if not same_kernel and kernels[i].name(
                kernels[i].NameBase_DEMANGLED
            ) == first_warmup_kernel.name(first_warmup_kernel.NameBase_DEMANGLED):
                warmup_end_idx = i
                break
        remove_warmup_kernels = kernels[warmup_end_idx:]
        # Remove duplicate kernels
        remove_dupe_kernels = []
        for kernel in remove_warmup_kernels:
            dupe = False
            for other_kernel in remove_dupe_kernels:
                if other_kernel.name(kernel.NameBase_DEMANGLED) == kernel.name(
                    kernel.NameBase_DEMANGLED
                ):
                    dupe = True
            if not dupe:
                remove_dupe_kernels.append(kernel)

        # Dictionary for metric values
        data_dict = defaultdict(list)
        # Matches everything between "<>"
        regex_str = r".*?\<(.*)\>.*"
        # For assertion
        first_kernel_metric_count = len(remove_dupe_kernels[0].metric_names())
        # Match kernels and add data
        for kernel in remove_dupe_kernels:
            kernel_name = kernel.name(kernel.NameBase_DEMANGLED)
            kernel_match = re.search(regex_str, kernel_name).group(1)
            ncu_side_kernel = kernel_name.replace(kernel_match, "").replace(" ", "")
            matches = []
            for other_kernel in cpu_side_kernels:
                k_match = re.search(regex_str, other_kernel).group(1)
                cpu_side_kernel = other_kernel.replace(k_match, "").replace(" ", "")
                if ncu_side_kernel == cpu_side_kernel:
                    matches.append(cpu_side_kernels[other_kernel])
                    # Remove entry since it should not be re-usable
                    cpu_side_kernels.pop(other_kernel)
                    break
            if len(matches) == 0:
                print(f"Could not match {kernel_name}")
                continue
            # Add metrics from NCU side
            data_dict["node"].append(matches[0])
            metrics = [kernel[name] for name in kernel.metric_names()]
            # Undefined behavior if this isn't true. We assume all kernels have same amount of metrics in the same order.
            assert len(metrics) == first_kernel_metric_count
            for metric in metrics:
                data_dict[metric.name()].append(metric.value())

        # Create NCU df
        ncu_df = pd.DataFrame.from_dict(data_dict)
        ncu_df["profile"] = ncu_hash
        ncu_df.set_index(["node", "profile"], inplace=True)

        # Get subset of metrics
        if chosen_metrics:
            ncu_df = ncu_df[chosen_metrics]

        # Append df to list
        ncu_df_list.append(ncu_df)

    # First join NCU dfs
    ncu_df = pd.concat(ncu_df_list)

    # Join Thicket and NCU dfs
    th.dataframe = th.dataframe.join(
        ncu_df,
        how="outer",
        sort=True,
        lsuffix="_left",
        rsuffix="_right",
    )



# Copyright 2022 Lawrence Livermore National Security, LLC and other
# Thicket Project Developers. See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT

from collections import OrderedDict


def verify_sorted_profile(thicket_component):
    """Assertion to check if profiles are sorted in a thicket dataframe

    Arguments:
        thicket_component (DataFrame): component of thicket to check
    """
    profile_index_values = list(
        OrderedDict.fromkeys(thicket_component.index.get_level_values("profile"))
    )
    if profile_index_values != sorted(profile_index_values):
        raise ValueError(
            "The profiles in this dataframe must be sorted. Try 'pandas.DataFrame.sort_index'."
        )


def verify_thicket_structures(thicket_component, columns=[], index=[]):
    """Assertion for missing input requirements to execute thicket functions.

    Arguments:
        thicket_components (DataFrame): component of thicket to check
        columns (list): list of columns to check for in the component
        index (list): list of index levels to check for in the component

    Returns:
        Raises an error if any columns or index levels are not in component, continues
            program if all columns and index levels are in component
    """
    if not isinstance(columns, list):
        raise RuntimeError("columns= must be specified as a list")
    if not isinstance(index, list):
        raise RuntimeError("index= must be specified as a list")

    # collect component columns and index
    component_columns = thicket_component.columns.tolist()
    component_index = thicket_component.index.names

    # check existence of columns and index
    column_result = all(elem in component_columns for elem in columns)
    index_result = all(elem in component_index for elem in index)

    # store missing columns or index if they exist
    missing_columns = list(set(columns).difference(component_columns))
    missing_index = list(set(index).difference(component_index))

    if not column_result and not index_result:
        raise RuntimeError(
            "Specified column(s) not found: "
            + str(missing_columns)
            + "\nMissing index level(s): "
            + str(missing_index)
            + " required for the function"
        )
    elif not column_result and index_result:
        raise RuntimeError("Specified column(s) not found: " + str(missing_columns))
    elif column_result and not index_result:
        raise RuntimeError(
            "Missing index level(s): "
            + str(missing_index)
            + " required for the function"
        )
